import gc
import warnings
from typing import Union

import numpy as np
from matplotlib import pyplot as plt
from scipy import signal
import xarray as xr

from mlair.helpers import to_list, TimeTrackingWrapper


class FIRFilter:

    def __init__(self, data, fs, order, cutoff, window, dim):

        filtered = []
        h = []
        for i in range(len(order)):
            fi, hi = self.apply_fir_filter(data, fs, order[i], cutoff_low=cutoff[i][0], cutoff_high=cutoff[i][1],
                                           window=window, dim=dim)
            filtered.append(fi)
            h.append(hi)

        self._filtered = filtered
        self._h = h

    def filter_coefficients(self):
        return self._h

    def filtered_data(self):
        return self._filtered
        #
        # y, h = fir_filter(station_data.values.flatten(), fs, order[0], cutoff_low=cutoff[0][0], cutoff_high=cutoff[0][1],
        #                   window=window)
        # filtered = xr.ones_like(station_data) * y.reshape(station_data.values.shape)
        # # band pass
        # y_band, h_band = fir_filter(station_data.values.flatten(), fs, order[1], cutoff_low=cutoff[1][0],
        #                             cutoff_high=cutoff[1][1], window=window)
        # filtered_band = xr.ones_like(station_data) * y_band.reshape(station_data.values.shape)
        # # band pass 2
        # y_band_2, h_band_2 = fir_filter(station_data.values.flatten(), fs, order[2], cutoff_low=cutoff[2][0],
        #                                 cutoff_high=cutoff[2][1], window=window)
        # filtered_band_2 = xr.ones_like(station_data) * y_band_2.reshape(station_data.values.shape)
        # # high pass
        # y_high, h_high = fir_filter(station_data.values.flatten(), fs, order[3], cutoff_low=cutoff[3][0],
        #                             cutoff_high=cutoff[3][1], window=window)
        # filtered_high = xr.ones_like(station_data) * y_high.reshape(station_data.values.shape)

    def apply_fir_filter(self, data, fs, order=5, cutoff_low=None, cutoff_high=None, window="hamming", dim="variables"):

        # create fir filter coeffs
        cutoff = []
        if cutoff_low is not None:
            cutoff += [cutoff_low]
        if cutoff_high is not None:
            cutoff += [cutoff_high]
        if len(cutoff) == 2:
            filter_type = "bandpass"
        elif len(cutoff) == 1 and cutoff_low is not None:
            filter_type = "highpass"
        elif len(cutoff) == 1 and cutoff_high is not None:
            filter_type = "lowpass"
        else:
            raise ValueError("Please provide either cutoff_low or cutoff_high.")
        h = signal.firwin(order, cutoff, pass_zero=filter_type, fs=fs, window=window)

        # filter data
        filtered = xr.ones_like(data)
        for var in data.coords[dim]:
            d = data.sel({dim: var}).values.flatten()
            y = signal.lfilter(h, 1., d)
            filtered.loc[{dim: var}] = y
        return filtered, h


class KolmogorovZurbenkoBaseClass:

    def __init__(self, df, wl, itr, is_child=False, filter_dim="window"):
        """
        It create the variables associate with the Kolmogorov-Zurbenko-filter.

        Args:
            df(pd.DataFrame, None): time series of a variable
            wl(list of int): window length
            itr(list of int): number of iteration
        """
        self.df = df
        self.filter_dim = filter_dim
        self.wl = to_list(wl)
        self.itr = to_list(itr)
        if abs(len(self.wl) - len(self.itr)) > 0:
            raise ValueError("Length of lists for wl and itr must agree!")
        self._isChild = is_child
        self.child = self.set_child()
        self.type = type(self).__name__

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoBaseClass(None, self.wl[1:], self.itr[1:], True, self.filter_dim)
        else:
            return None

    def kz_filter(self, df, m, k):
        pass

    def spectral_calc(self):
        df_start = self.df
        kz = self.kz_filter(df_start, self.wl[0], self.itr[0])
        filtered = self.subtract(df_start, kz)
        # case I: no child avail -> return kz and remaining
        if self.child is None:
            return [kz, filtered]
        # case II: has child -> return current kz and all child results
        else:
            self.child.df = filtered
            kz_next = self.child.spectral_calc()
            return [kz] + kz_next

    @staticmethod
    def subtract(minuend, subtrahend):
        try:  # pandas implementation
            return minuend.sub(subtrahend, axis=0)
        except AttributeError:  # general implementation
            return minuend - subtrahend

    def run(self):
        return self.spectral_calc()

    def transfer_function(self):
        m = self.wl[0]
        k = self.itr[0]
        omega = np.linspace(0.00001, 0.15, 5000)
        return omega, (np.sin(m * np.pi * omega) / (m * np.sin(np.pi * omega))) ** (2 * k)

    def omega_null(self, alpha=0.5):
        a = np.sqrt(6) / np.pi
        b = 1 / (2 * np.array(self.itr))
        c = 1 - alpha ** b
        d = np.array(self.wl) ** 2 - alpha ** b
        return a * np.sqrt(c / d)

    def period_null(self, alpha=0.5):
        return 1. / self.omega_null(alpha)

    def period_null_days(self, alpha=0.5):
        return self.period_null(alpha) / 24.

    def plot_transfer_function(self, fig=None, name=None):
        if fig is None:
            fig = plt.figure()
        omega, transfer_function = self.transfer_function()
        if self.child is not None:
            transfer_function_child = self.child.plot_transfer_function(fig)
        else:
            transfer_function_child = transfer_function * 0
        plt.semilogx(omega, transfer_function - transfer_function_child,
                     label="m={:3.0f}, k={:3.0f}, T={:6.2f}d".format(self.wl[0],
                                                                     self.itr[0],
                                                                     self.period_null_days()))
        plt.axvline(x=self.omega_null())
        if not self._isChild:
            locs, labels = plt.xticks()
            plt.xticks(locs, np.round(1. / (locs * 24), 1))
            plt.xlim([0.00001, 0.15])
            plt.legend()
            if name is None:
                plt.show()
            else:
                plt.savefig(name)
        else:
            return transfer_function


class KolmogorovZurbenkoFilterMovingWindow(KolmogorovZurbenkoBaseClass):

    def __init__(self, df, wl: Union[list, int], itr: Union[list, int], is_child=False, filter_dim="window",
                 method="mean", percentile=0.5):
        """
        It create the variables associate with the KolmogorovZurbenkoFilterMovingWindow class.

        Args:
            df(pd.DataFrame, xr.DataArray): time series of a variable
            wl: window length
            itr: number of iteration
        """
        self.valid_methods = ["mean", "percentile", "median", "max", "min"]
        if method not in self.valid_methods:
            raise ValueError("Method '{}' is not supported. Please select from [{}].".format(
                method, ", ".join(self.valid_methods)))
        else:
            self.method = method
            if percentile > 1 or percentile < 0:
                raise ValueError("Percentile must be in range [0, 1]. Given was {}!".format(percentile))
            else:
                self.percentile = percentile
        super().__init__(df, wl, itr, is_child, filter_dim)

    def set_child(self):
        if len(self.wl) > 1:
            return KolmogorovZurbenkoFilterMovingWindow(self.df, self.wl[1:], self.itr[1:], is_child=True,
                                                        filter_dim=self.filter_dim, method=self.method,
                                                        percentile=self.percentile)
        else:
            return None

    @TimeTrackingWrapper
    def kz_filter_new(self, df, wl, itr):
        """
        It passes the low frequency time series.

        If filter method is from mean, max, min this method will call construct and rechunk before the actual
        calculation to improve performance. If filter method is either median or percentile this approach is not
        applicable and depending on the data and window size, this method can become slow.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        warnings.filterwarnings("ignore")
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": int(0.7 * wl),
                      "center": True,
                      self.filter_dim: wl}
            for i in np.arange(0, itr):
                print(i)
                rolling = df_itr.chunk().rolling(**kwargs)
                if self.method not in ["percentile", "median"]:
                    rolling = rolling.construct("construct").chunk("auto")
                if self.method == "median":
                    df_mv_avg_tmp = rolling.median()
                elif self.method == "percentile":
                    df_mv_avg_tmp = rolling.quantile(self.percentile)
                elif self.method == "max":
                    df_mv_avg_tmp = rolling.max("construct")
                elif self.method == "min":
                    df_mv_avg_tmp = rolling.min("construct")
                else:
                    df_mv_avg_tmp = rolling.mean("construct")
                df_itr = df_mv_avg_tmp.compute()
                del df_mv_avg_tmp, rolling
                gc.collect()
            return df_itr
        except ValueError:
            raise ValueError

    @TimeTrackingWrapper
    def kz_filter(self, df, wl, itr):
        """
        It passes the low frequency time series.

        Args:
             wl(int): a window length
             itr(int): a number of iteration
        """
        import warnings
        warnings.filterwarnings("ignore")
        df_itr = df.__deepcopy__()
        try:
            kwargs = {"min_periods": int(0.7 * wl),
                      "center": True,
                      self.filter_dim: wl}
            iter_vars = df_itr.coords["variables"].values
            for var in iter_vars:
                df_itr_var = df_itr.sel(variables=[var])
                for _ in np.arange(0, itr):
                    df_itr_var = df_itr_var.chunk()
                    rolling = df_itr_var.rolling(**kwargs)
                    if self.method == "median":
                        df_mv_avg_tmp = rolling.median()
                    elif self.method == "percentile":
                        df_mv_avg_tmp = rolling.quantile(self.percentile)
                    elif self.method == "max":
                        df_mv_avg_tmp = rolling.max()
                    elif self.method == "min":
                        df_mv_avg_tmp = rolling.min()
                    else:
                        df_mv_avg_tmp = rolling.mean()
                    df_itr_var = df_mv_avg_tmp.compute()
                df_itr.loc[{"variables": [var]}] = df_itr_var
            return df_itr
        except ValueError:
            raise ValueError
